/* 2005/03/23 */

/* A Link-layer indications program for wireless devices
   using RTNETLINK. */

/* PURPOSE:
	An IPv6 Router Solicitation is transmitted when a 
	host receives an association (sorry, no RSN) from
	802.11 wireless LAN..

	This is a hacked example. Code style has been ignored

	Its primary uses are in my experiments and for
	illustration purposes.

*/

/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to 
the GNU General Public License v2 or greater.

Please see README.txt for details. 
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/
/*************************************************************/




/*  GD: Note to self: 20040918 origin
    GD: Note to self: 20050110 evtest

     REF RFC 3549, for RTNETLINK msg  formats and types.


*/

/**************************************************************************/

/*ipv6 socket headers  AAA */
#include <sys/socket.h>
#include <netinet/in.h>
/* interface request headers  BBB */
#include <sys/ioctl.h>
#include <net/if.h>
/* extra ifreq headers*/
#include <string.h>

/* rtnetlink socket includes */
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>


/* local diagnostic headers */
#include <stdio.h>
#include <stdlib.h>		/* exit(1)... */
#include <errno.h>

/* time information */
#include <time.h>
#include <sys/time.h>


#ifndef IFLA_WIRELESS
#define IFLA_WIRELESS 11	/* ???? */
#endif				/*IFLA_WIRELESS */

#ifndef _LINUX_IF_H
#define _LINUX_IF_H
#define _UNDEF_LINUX_IF_H
#endif				/*  _LINUX_IF_H */

#include <linux/wireless.h>

#ifdef _UNDEF_LINUX_IF_H
#undef _UNDEF_LINUX_IF_H
#undef _LINUX_IF_H
#endif				/* _UNDEF_LINUX_IF_H */

#include <sys/select.h>

#include "eventproc.h"
#include "inputs.h"
extern int send_rs(int sock, int ifindex, uint8_t macaddr[], int tentative,
		   struct in6_addr *src, struct in6_addr *dst, uint8_t *
		   prefix);


extern int bind_rtsock(uint32_t groups);
extern int open_v6sock(void);
extern int get_ifindex(int sock, const char *name);

extern int get_mac_rtnetlink(int sock, int ifindex, unsigned char *mac);
extern int get_linklocal_rtnetlink(int sock, int ifindex, unsigned char *lladdr);

extern int get_li_rtnetlink(int sock, struct link_status *state);


/**************************************************************************/


int main(int argc, char *argv[])
{
	int rtsock;
	int v6sock;
	/*char ifname[IFNAMSIZ]="eth1"; */
	char name[IFNAMSIZ + 1];

	int ifindex;
	uint8_t macaddr[6];
	uint8_t prefix[16];
	int rval = 0;
	int macaddrset = 0;
	struct in6_addr *src = NULL;
	struct in6_addr srcbuff;
	int usesrcaddr = 0;
	int tentative = 0;
	int cisco=0;
	struct timeval tv;
	struct tm *tm;
	struct link_status state;

    /***********************************************************************/
	fd_set rfds, efds;
    /***********************************************************************/

	memset(macaddr, 0, 6 * sizeof(uint8_t));
	memset(prefix, 0, 16 * sizeof(uint8_t));

	if ((argc > 1) && !strncmp("-C", argv[argc-1], 3)) {
		cisco = 1;
		argc--;
	}
	if ((argc > 2) && (argv[argc - 1][0] != '-')
	    && (argv[argc - 2][0] == '-')
	    && (argv[argc - 2][1] == 'L')) {
		sscanf(argv[argc - 1],
		       "%02hhx%02hhx:%02hhx%02hhx:%02hhx%02hhx:%02hhx%02hhx",
		       &prefix[0], &prefix[1], &prefix[2], &prefix[3],
		       &prefix[4], &prefix[5], &prefix[6], &prefix[7]);
		argc -= 2;
	}
	if ((argc > 2) && (argv[argc - 1][0] != '-')) {
		sscanf(argv[argc - 1],
		       "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
		       &macaddr[0], &macaddr[1], &macaddr[2], &macaddr[3],
		       &macaddr[4], &macaddr[5]);
		macaddrset = 1;
		argc--;
	}
	if ((argc == 4) && !strncmp("-t", argv[3], 3)) {
		tentative = 1;
		argc = 3;
	}

	if ((argc == 3) && !strncmp("-s", argv[2], 3)) {
		usesrcaddr = 1;
		argc = 2;
	}
	if (argc != 2) {
		fprintf(stderr,
			"Usage: %s <interface> [-s  [-t] [<ethmac>] [-L <prefix>] [-C]\n"
			"\t-s:  Use unicast source address\n"
			"\t-t:  Use tentative link-layer addresses\n"
			"\t-C:  Cisco Aironet Driver used\n" 
			"\t<ethmac>:  XX:XX:XX:XX:XX:XX\n"
			"\t<prefix>:  XXXX:XXXX:XXXX:XXXX\n", argv[0]);
		exit(1);
	}
	strncpy(name, argv[1], IFNAMSIZ);




    /***********************************************************************/

	if ((rtsock = bind_rtsock(RTMGRP_LINK)) < 0) {
		fprintf(stderr, "unable to bind netlink socket\n");
		exit(1);
	}

	if ((v6sock = open_v6sock()) < 0) {
		fprintf(stderr, "unable to get ipv6 socket\n");
		exit(1);
	}

	if ((ifindex = get_ifindex(v6sock, name)) < 0) {
		fprintf(stderr, "unable to get ifindex\n");
		exit(1);
	}
	if (!macaddrset && get_mac_rtnetlink(rtsock, ifindex, macaddr) < 0) {
		fprintf(stderr,"mac address setup failed on %s\n", name);
		return -1;
	}
	fprintf(stderr,"MAC is %2.2hhx:%2.2hhx:%2.2hhx:%2.2hhx:%2.2hhx:%2hhx\n",
	     macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4],
	     macaddr[5]);

	if (usesrcaddr) {
		src = &srcbuff;
		if (get_linklocal_rtnetlink(rtsock, ifindex, (unsigned char *)
				src) <= 0) {
			fprintf(stderr,"lladdr setup failed %s\n", name);
			return -1;
		}
		fprintf(stderr,"Link-local address: "
			"%hx:%hx:%hx:%hx:%hx:%hx:%hx:%hx\n",
			ntohs(src->s6_addr16[0]),
			ntohs(src->s6_addr16[1]),
			ntohs(src->s6_addr16[2]),
			ntohs(src->s6_addr16[3]),
			ntohs(src->s6_addr16[4]),
			ntohs(src->s6_addr16[5]),
			ntohs(src->s6_addr16[6]),
			ntohs(src->s6_addr16[7]));
	}

	memcpy(&state.source, &srcbuff, sizeof(struct in6_addr));
        memcpy(&state.landmark, &prefix, sizeof(struct in6_addr));
        state.landplen=128;
        state.ifindex=ifindex;
        state.tentative=tentative;
        state.v6sock=0;
        state.rasock=0;
	state.cisco=cisco;
        memcpy(state.macaddr, macaddr, 6);	

    /***********************************************************************/
	FD_ZERO(&efds);
	FD_SET(rtsock, &efds);

	while (1) {
		FD_ZERO(&rfds);
		FD_SET(rtsock, &rfds);
	 	if((rval=select(rtsock+1, &rfds, NULL, &efds, NULL))<0){
			break;
		}
		if(FD_ISSET(rtsock, &efds)){
			break;
		}
		fprintf(stderr,"Socket readable\n");
		while((rval = get_li_rtnetlink(rtsock, &state)) >= 0){
			if(!rval)
				break;
			/*  0 return ifindex of link up??? */
			gettimeofday(&tv, NULL);
			tm = localtime(&tv.tv_sec);
			fprintf(stderr,"link up event: %d:%2.2d:%2.2d.%6.6lu\n",
			       tm->tm_hour, tm->tm_min, tm->tm_sec,
			       tv.tv_usec);
			send_rs(v6sock, ifindex, macaddr, tentative, src,
				NULL, prefix);

		}
	}

	return 0;
}



#if 0
int process_link_event(struct nlmsghdr *nlm)
{
	struct ifinfomsg *infom;
	int ml;
	unsigned int rtal;
	struct rtattr *rta;
	unsigned char *rtap;	/* RTA_PAYLOAD */
	struct sockaddr *sll;

	struct iw_event *iwe;
	int rval = 0;
	unsigned char noaddr[8];

	memset(&noaddr, 0, 8);

	infom = NLMSG_DATA(nlm);
	fprintf(stderr, "nll: %d nlf:%x ifif:%d ifit:%d ifind:%d flags: %x change %x \n",
           nlm->nlmsg_len, nlm->nlmsg_flags,
	   infom->ifi_family, infom->ifi_type,
	   infom->ifi_index, infom->ifi_flags, infom->ifi_change); 
	ml = nlm->nlmsg_len;
	rta = IFLA_RTA(infom);

	while (RTA_OK(rta, ml)) {
		/*  fprintf(stderr, "rok:%p  rtal: %d  ml%d ",
		   rta, rta->rta_len, rta->rta_type); */

		rtap = RTA_DATA(rta);
		rtal = RTA_PAYLOAD(rta);
		switch (rta->rta_type) {
		case IFLA_WIRELESS:
			iwe = (struct iw_event *) rtap;
			fprintf(stderr,"WIRELESS %hx (l:%d) ", iwe->cmd, iwe->len);
			switch (iwe->cmd) {
			case SIOCGIWAP:
				/* always family 1 (sockaddr) */
				sll = &iwe->u.ap_addr;
				fprintf(stderr,"SIOCGIWAP ");
				if (!memcmp(&sll->sa_data, noaddr, 6)) {
					/* disassociate */
					fprintf(stderr,"disassoc ");
				} else {
					fprintf(stderr,"associate "
					       "%.2hhx:%.2hhx:"
					       "%.2hhx:%.2hhx:"
					       "%.2hhx:%.2hhx",
					       sll->sa_data[0],
					       sll->sa_data[1],
					       sll->sa_data[2],
					       sll->sa_data[3],
					       sll->sa_data[4],
					       sll->sa_data[5]);
					rval = infom->ifi_index;
				}
				break;
			case SIOCSIWESSID:
				fprintf(stderr," SIOCSIWESSID");
				break;
			default:
				break;
			}
			break;
		default:
			fprintf(stderr,"type:%d ", rta->rta_type);
			break;
		}
		rta = RTA_NEXT(rta, ml);
	}

	return rval;



}

int get_li_rtnetlink(int rtnetlink_socket, int ifindex)
{
	int rval = 0;
	int recvsize;
	struct nlmsghdr *nlm;
	struct {
		struct nlmsghdr nh;
		char attrbuf[1024];
	} req;

	memset(&req, 0, sizeof(req));


	if ((recvsize = recv(rtnetlink_socket, &req, sizeof(req), 0)) > 0) {


		nlm = (struct nlmsghdr *) &req;
		while (NLMSG_OK(nlm, recvsize)) {
			switch (nlm->nlmsg_type) {
			case NLMSG_ERROR:
				fprintf(stderr,"error: %d (%s)\n",
				       ((struct nlmsgerr *)
					NLMSG_DATA(&req))->error,
				       strerror(-((struct nlmsgerr *)
						  NLMSG_DATA(&req))->
						error));
				exit(1);
				break;

			case RTM_NEWLINK:
				fprintf(stderr,"rtm_newlink ");

				rval = process_link_event(nlm);

				fprintf(stderr,"\n");
				break;
			case RTM_DELLINK:
				fprintf(stderr,"rtm_dellink\n");

				return 0;

			case NLMSG_DONE:
				return rval;
				break;
			default:
				fprintf(stderr,"NLMSG Type: %d recvsize=%d\n",
				       req.nh.nlmsg_type, recvsize);
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if (recvsize < 0) {
		return -1;
	}
	return rval;
}

#endif /* 0 */



